/*!
 ************************************************************************
 * \file adaptive_filter.c
 *
 * \brief
 *    adaptive filter routines
 *
 * \author
 *  Yuri Vatis               <vatis@tnt.uni-hannover.de>    \n
 *
 * This software may be used for research purposes only.
 * Universitaet Hannover would also appreciate that all those using the 
 * software or any extension of it in any way comply with the 
 * following condition:
 *    Including in any technical report, conference or journal paper
 *    produced which uses the software or any extension of it in
 *    any way, at least one of bibliographic references:
 *    Yuri Vatis, Bernd Edler, Dieu Thanh Nguyen, Jrn Ostermann,
 *    "Motion-and Aliasing-compensated Prediction using a two-dimensional non-separable 
 *    Adaptive Wiener Interpolation Filter",
 *		Proc. ICIP 2005, IEEE International Conference on Image Processing, Genova, Italy, September 2005.
 *    Yuri Vatis, Bernd Edler, Ingolf Wassermann, Dieu Thanh Nguyen, Jrn Ostermann,
 *    "Coding of Coefficients of two-dimensional non-separable Adaptive Wiener Interpolation Filter", 
 *		Proc. VCIP 2005, SPIE Visual Communication & Image Processing, Beijing, China, July 2005.
 ************************************************************************
 */

   

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "global.h"
#include "adaptive_filter.h"
#include "memalloc.h"
#include "malloc.h"
#include "defines.h"
#include "mbuffer.h"
#include "image.h"
#include "vlc.h"

#ifdef ADAPTIVE_FILTER
static int TwoDEquationPattern[15][SQR_FILTER]  =		// get equation number for symmetric filter coefficients
{
	{ 0,  1,  2,  3,  4,  5,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,        // a_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  1,  2,  2,  1,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,        // b_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 5,  4,  3,  2,  1,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,        // c_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  1,  2,  3,  4,  5,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,        // d_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  1,  2,  3,  4,  5,
	1,  6,  7,  8,  9, 10,
	2,  7, 11, 12, 13, 14,        // e_pos
	3,  8, 12, 15, 16, 17,
	4,  9, 13, 16, 18, 19,
	5, 10, 14, 17, 19, 20	},
	{ 0,  1,  2,  2,  1,  0,
	3,  4,  5,  5,  4,  3,
	6,  7,  8,  8,  7,  6,        // f_pos
	9, 10, 11, 11, 10,  9,
	12, 13, 14, 14, 13, 12,
	15, 16, 17, 17, 16, 15  },
	{ 5,  4,  3,  2,  1,  0,
	10,  9,  8,  7,  6,  1,
	14, 13, 12, 11,  7,  2,        // g_pos
	17, 16, 15, 12,  8,  3,
	19, 18, 16, 13,  9,  4,
	20, 19, 17, 14, 10,  5	},
	{ 0,  1,  2,  2,  1,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,        // h_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  3,  6,  9, 12, 15,
	1,  4,  7, 10, 13, 16,
	2,  5,  8, 11, 14, 17,        // i_pos
	2,  5,  8, 11, 14, 17,
	1,  4,  7, 10, 13, 16,
	0,  3,  6,  9, 12, 15  },
	{ 0,  1,  2,  2,  1,  0,
	1,  3,  4,  4,  3,  1,
	2,  4,  5,  5,  4,  2,   			// j_pos
	2,  4,  5,  5,  4,  2,
	1,  3,  4,  4,  3,  1,
	0,  1,  2,  2,  1,  0  },
	{15, 12,  9,  6,  3,  0,
	16, 13, 10,  7,  4,  1,
	17, 14, 11,  8,  5,  2,   			// k_pos
	17, 14, 11,  8,  5,  2,
	16, 13, 10,  7,  4,  1,
	15, 12,  9,  6,  3,  0  },
	{ 5,  4,  3,  2,  1,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,         // l_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 5, 10, 14, 17, 19, 20,
	4,  9, 13, 16, 18, 19,
	3,  8, 12, 15, 16, 17,   			// m_pos
	2,  7, 11, 12, 13, 14,
	1,  6,  7,  8,  9, 10,
	0,  1,  2,  3,  4,  5	},
	{15, 16, 17, 17, 16, 15,
	12, 13, 14, 14, 13, 12,
	9, 10, 11, 11, 10,  9,   			// n_pos
	6,  7,  8,  8,  7,  6,
	3,  4,  5,  5,  4,  3,
	0,  1,  2,  2,  1,  0  },
	{20, 19, 17, 14, 10,  5,
	19, 18, 16, 13,  9,  4,
	17, 16, 15, 12,  8,  3,   			// o_pos
	14, 13, 12, 11,  7,  2,
	10,  9,  8,  7,  6,  1,
	5,  4,  3,  2,  1,  0	}
};

static double STANDARD_2D_FILTER[15][SQR_FILTER] = {
{ 0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   52.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // a_pos
{ 0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	1.0/32.0  ,   -5.0/32.0  ,   20.0/32.0  ,   20.0/32.0  ,   -5.0/32.0  ,   1.0/32.0  ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // b_pos
{ 0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   52.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // c_pos
{ 0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // d_pos
{ 0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
	}, // e_pos
{ 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
 52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
 20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // f_pos
{ 0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0
}, // g_pos
{ 0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0
}, // h_pos
{ 1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
 20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // i_pos
{ 1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0,
 -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
 20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
 20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
 -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
  1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0
}, // j_pos
{ 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
 20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // i_pos
{ 0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // l_pos
{ 0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // m_pos
{ 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
 20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
 52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // n_pos
{ 0.0       ,    0.0       ,    0.0     ,      1.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0     ,     -5.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0     ,     20.0/64.0  ,    0.0       ,   0.0       ,
	1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0,     40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
	0.0       ,    0.0       ,    0.0     ,     -5.0/64.0  ,    0.0       ,   0.0       ,
	0.0       ,    0.0       ,    0.0     ,      1.0/64.0  ,    0.0       ,   0.0
	} // o_pos
};

static int FILTER_NEW_SUB_POS[SQR_FILTER] = {        a_pos, b_pos, a_pos,
	a_pos, e_pos, f_pos, e_pos,
	b_pos, f_pos, j_pos, f_pos,
	a_pos, e_pos, f_pos, e_pos };
static int TwoDSymmetricPattern[15][SQR_FILTER]  =		// get one filter from another one, if symmetry properties are used (used e.g. in ExtendFilterCoefficients)
{
	{ 0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		12, 13, 14, 15, 16, 17,        // a_pos  - not used actually
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0  },
	{ 0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		12, 13, 14, 14, 13, 12,        // b_pos
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0  },
	{ 0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		17, 16, 15, 14, 13, 12,        // c_pos
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0  },
	{ 0,  0, 12,  0,  0,  0,
		0,  0, 13,  0,  0,  0,
		0,  0, 14,  0,  0,  0,        // d_pos
		0,  0, 15,  0,  0,  0,
		0,  0, 16,  0,  0,  0,
		0,  0, 17,  0,  0,  0  },
	{ 0,  1,  2,  3,  4,  5,
		1,  7,  8,  9, 10, 11,
		2,  8, 14, 15, 16, 17,        // e_pos
		3,  9, 15, 21, 22, 23,
		4, 10, 16, 22, 28, 29,
		5, 11, 17, 23, 29, 35  },
	{ 0,  1,  2,  2,  1,  0,
		6,  7,  8,  8,  7,  6,
		12, 13, 14, 14, 13, 12,        // f_pos
		18, 19, 20, 20, 19, 18,
		24, 25, 26, 26, 25, 24,
		30, 31, 32, 32, 31, 30  },
	{ 5,  4,  3,  2,  1,  0,
		11, 10,  9,  8,  7,  1,
		17, 16, 15, 14,  8,  2,        // g_pos
		23, 22, 21, 15,  9,  3,
		29, 28, 22, 16, 10,  4,
		35, 29, 23, 17, 11,  5  },
	{ 0,  0, 12,  0,  0,  0,
		0,  0, 13,  0,  0,  0,
		0,  0, 14,  0,  0,  0,        // h_pos
		0,  0, 14,  0,  0,  0,
		0,  0, 13,  0,  0,  0,
		0,  0, 12,  0,  0,  0  },
	{ 0,  6, 12, 18, 24, 30,
		1,  7, 13, 19, 25, 31,
		2,  8, 14, 20, 26, 32,        // i_pos
		2,  8, 14, 20, 26, 32,
		1,  7, 13, 19, 25, 31,
		0,  6, 12, 18, 24, 30  },
	{ 0,  1,  2,  2,  1,  0,
		1,  7,  8,  8,  7,  1,
		2,  8, 14, 14,  8,  2,   			// j_pos
		2,  8, 14, 14,  8,  2,
		1,  7,  8,  8,  7,  1,
		0,  1,  2,  2,  1,  0  },
	{30, 24, 18, 12,  6,  0,
	 31, 25, 19, 13,  7,  1,
	 32, 26, 20, 14,  8,  2,   			// k_pos
	 32, 26, 20, 14,  8,  2,
	 31, 25, 19, 13,  7,  1,
	 30, 24, 18, 12,  6,  0  },
	{ 0,  0, 17,  0,  0,  0,
		0,  0, 16,  0,  0,  0,
		0,  0, 15,  0,  0,  0,         // l_pos
		0,  0, 14,  0,  0,  0,
		0,  0, 13,  0,  0,  0,
		0,  0, 12,  0,  0,  0  },
	{ 5, 11, 17, 23, 29, 35,
		4, 10, 16, 22, 28, 29,
		3,  9, 15, 21, 22, 23,        // m_pos
		2,  8, 14, 15, 16, 17,
		1,  7,  8,  9, 10, 11,
		0,  1,  2,  3,  4,  5  },
	{30, 31, 32, 32, 31, 30,
   24, 25, 26, 26, 25, 24,
	 18, 19, 20, 20, 19, 18,   			// n_pos
	 12, 13, 14, 14, 13, 12,
	 	6,  7,  8,  8,  7,  6,
		0,  1,  2,  2,  1,  0  },
	{35, 29, 23, 17, 11,  5,
	 29, 28, 22, 16, 10,  4,
	 23, 22, 21, 15,  9,  3,   			// o_pos
	 17, 16, 15, 14,  8,  2,
	 11, 10,  9,  8,  7,  1,
		5,  4,  3,  2,  1,  0	}
};
	int DiffQFilterCoef[15][SQR_FILTER];				// differences to be transmitted
	int POS_EQUATION_NUMBER[15];								// number of different coefficietns for each sub-pel position
	int FilterFlag[15];													// Flags defining if Filter at the particular position calculated
	int	UseAllSubpelPositions;									// 1 if FilterFlag for all independent positions is 1
	int SubpelPositionsPattern;
	int SymmetryPosition[15] = {1,1,0,0,1,1,0,0,0,1,0,0,0,0,0};	// if 0, the position is copied from another one
  int Is1DPosition[15] = {1,1,1,1,0,0,0,1,0,0,0,1,0,0,0}; // if 0, the position is a 2D one
	int FILTER_SIZE = 6;		// is not checked for other values
	int FILTER_OFFSET;
	int NumberOfQBits;                              // number of bits used for quantization of the filter coefficients



	static double NxN_Matrix[15][SQR_FILTER][SQR_FILTER];			// SQR(PEL)-1 X SQR(FILTER_SIZE) X SQR(FILTER_SIZE) - matrix
	static double N_Vector[15][SQR_FILTER];										// SQR(PEL)-1 X SQR(FILTER_SIZE) - vector
	static double FilterCoef[15][SQR_FILTER];									// estimated filter coefficients
	static int    QFilterCoef[15][SQR_FILTER];								// SQR(PEL)-1 X SQR(FILTER_SIZE) - quantized filter coefficients
//	static double LastFilterCoef[15][SQR_FILTER];							// filter used for previous frame
	static double PredFilterCoef[15][SQR_FILTER];							// predicted filter coefficients
	static double	DiffFilterCoef[15][SQR_FILTER];							// differences
	static double CalculatedFilter[15][SQR_FILTER];						// Filter for different positions
	static int **MBRateStorage;													// storage for number of bits per macroblock for part. mode coded with standard filter
	static int *BitsPerFrameCounter;													// storage for number of bits per macroblock for part. mode coded with standard filter
	static int number_of_macroblocks;
	static int current_poc;
/*!
 ************************************************************************
 * \brief
 *    init adaptive filter structures
 ************************************************************************
 */
void InitAdaptiveFilter(InputParameters *input)
{
	int block_x_nr;
	int block_y_nr;
	if(!input->UseAdaptiveFilter)
	{
		return;
	}
	NumberOfQBits = DEFAULT_QUANT;
	if(NumberOfQBits < 5 || NumberOfQBits > 16)
	{
		NumberOfQBits=DEFAULT_QUANT;
		printf("setting NumberOfQBits to default value...%d\n", NumberOfQBits);
	}
	FILTER_OFFSET = (FILTER_SIZE)/2-1;
	POS_EQUATION_NUMBER[a_pos] =  6;
	POS_EQUATION_NUMBER[b_pos] =  3;
	POS_EQUATION_NUMBER[c_pos] =  6;
	
	POS_EQUATION_NUMBER[d_pos] =  6;
	POS_EQUATION_NUMBER[e_pos] = 21;
	POS_EQUATION_NUMBER[f_pos] = 18;
	POS_EQUATION_NUMBER[g_pos] = 21;
	
	POS_EQUATION_NUMBER[h_pos] =  3;
	POS_EQUATION_NUMBER[i_pos] = 18;
	POS_EQUATION_NUMBER[j_pos] =  6;
	POS_EQUATION_NUMBER[k_pos] = 18;
	
	POS_EQUATION_NUMBER[l_pos] =  6;
	POS_EQUATION_NUMBER[m_pos] = 21;
	POS_EQUATION_NUMBER[n_pos] = 18;
	POS_EQUATION_NUMBER[o_pos] = 21;

	number_of_macroblocks = img->width*img->height/256;
	BitsPerFrameCounter = calloc(number_of_macroblocks, sizeof(int));
	get_mem2Dint(&MBRateStorage, number_of_macroblocks, MAXMODE);
	block_y_nr = img->height/4;
	block_x_nr = img->width/4;
	BitsPerFrameCounter = calloc(number_of_macroblocks, sizeof(int));
	get_mem2Dint(&MBRateStorage, number_of_macroblocks, MAXMODE);
	current_poc=-1;
}

/*!
 ************************************************************************
 * \brief
 *    free adaptive filter structure
 ************************************************************************
 */
void FreeAdaptiveFilter()
{
	int block_x_nr;
	int block_y_nr;
	if(!input->UseAdaptiveFilter)
		return;
	if(MBRateStorage)
		free_mem2Dint(MBRateStorage);
	if(BitsPerFrameCounter)
		free(BitsPerFrameCounter);
	block_y_nr = img->height/4;
	block_x_nr = img->width/4;
}


/*!
 ************************************************************************
 * \brief
 *    resets all filter data
 ************************************************************************
 */
void ResetAdaptiveFilter()
{
	int i, j, k;
	if(!UseAdaptiveFilterForCurrentFrame())
	{
		return;
	}
	else
	{
		for(k = 0; k < 15; k++)
		{
			for(i = 0; i < SQR_FILTER; i++)
			{
				for(j = 0; j < SQR_FILTER; j++)
				{
					NxN_Matrix[k][i][j] = 0.0;
				}
				N_Vector[k][i]       = 0.0;
				FilterCoef[k][i]     = 0.0;
				QFilterCoef[k][i]		 = 0;
				PredFilterCoef[k][i] = 0.0;
			}
			FilterFlag[k] = 0;
		}
	}
	NumberOfQBits = DEFAULT_QUANT;
}


/*!
 ************************************************************************
 * \brief
 *    return if adaptive filter is used for current frame
 ************************************************************************
*/
int UseAdaptiveFilterForCurrentFrame(void)
{
	if((img->type == P_SLICE || img->type == B_SLICE) && input->UseAdaptiveFilter)
		return 1;
	else
		return 0;
}


/*!
 ************************************************************************
 * \brief
 *    sets necessary filter data for current macroblock
 ************************************************************************
 */
void SetAdaptiveFilter(void)
{
	Macroblock *currMB = &img->mb_data[img->current_mb_nr];
	int i, j, x, y, xj, yi;
	float b_scaling=1.0;
	int run=0, rerun=0;
	int equation=0;
	int mvx_subF=0, mvy_subF=0;
	int mvx_subB=0, mvy_subB=0;
	int mvx=0,  mvy=0;
	int mvxF=0, mvyF=0;
	int mvxB=0, mvyB=0;
	int sub_pos=0;
	int sub_posF=0;
	int sub_posB=0;
	int ref_frame=0;
	int ref_frameF=-1;
	int ref_frameB=-1;
	int subblock=0;
	int filter_pos_x1, filter_pos_y1, filter_pos_x2, filter_pos_y2;
	int x_orig, y_orig;       // absolute distances from current macroblock
	int h1, h2, h3;
	int *new_filter_pos, *new_eq, *new_sub_pos;
	int number_equation;
	imgpel **RecY=NULL;
	imgpel **RecYF=NULL;
	imgpel **RecYB=NULL;
	new_filter_pos = &h1;
	new_eq = &h2;
	new_sub_pos = &h3;
#ifdef EIGHTH_PEL
	if(img->mv_res)
	{
		SetAdaptiveFilterEighthpel();
		return;
	}
#endif
	if(IS_INTRA(currMB)) //intra macroblocks are not used for calculation of the filter coeffs.
		return;

	x_orig = MB_BLOCK_SIZE*(img->block_x/4);
	y_orig = MB_BLOCK_SIZE*(img->block_y/4);
	for(subblock = 0; subblock < 16; subblock++)
	{
		x = x_orig+4*(subblock%4);
		y = y_orig+4*(subblock/4);
		mvxF = enc_picture->mv[LIST_0][y/4][x/4][0];
		mvyF = enc_picture->mv[LIST_0][y/4][x/4][1];
		if(mvxF >= 0)
			mvx_subF = mvxF%4;      // x-sub-coordinate in a 4x4block
		else
			mvx_subF = (4-abs(mvxF)%4)%4;
		if(mvyF >= 0)
			mvy_subF = mvyF%4;      // y-sub-coordinate in a 4x4block
		else
			mvy_subF = (4-abs(mvyF)%4)%4;

		sub_posF = mvx_subF + 4*mvy_subF-1;    // pos 0..14 in a 4x4 block
		ref_frameF = enc_picture->ref_idx[LIST_0][y/4][x/4];
		if(ref_frameF != -1)
			RecYF  = listX[LIST_0][ref_frameF]->imgY;
		if(img->type == B_SLICE)
		{
			mvxB= enc_picture->mv[LIST_1][y/4][x/4][0];
			mvyB= enc_picture->mv[LIST_1][y/4][x/4][1];
			if(mvxB >= 0)
				mvx_subB = mvxB%4;      // x-sub-coordinate in a 4x4block
			else
				mvx_subB = (4-abs(mvxB)%4)%4;
			if(mvyB >= 0)
				mvy_subB = mvyB%4;      // y-sub-coordinate in a 4x4block
			else
				mvy_subB = (4-abs(mvyB)%4)%4;
	
			sub_posB = mvx_subB + 4*mvy_subB-1;    // pos 0..14 in a 4x4 block
			ref_frameB = enc_picture->ref_idx[LIST_1][y/4][x/4];
			if(ref_frameB != -1)
				RecYB  = listX[LIST_1][ref_frameB]->imgY;
			rerun = 1;
    }
		else
			rerun = 0;
		if(ref_frameB != -1 && ref_frameF != -1)
			b_scaling = 0.5;
		else
			b_scaling = 1.0;
		for(run = 0; run <= rerun; run++)
		{
			if(run == 0)
			{
				sub_pos = sub_posF;
				mvx = mvxF;
				mvy = mvyF;
				RecY = RecYF;
				ref_frame = ref_frameF;
			}
			else
			{
				sub_pos = sub_posB;
				mvx = mvxB;
				mvy = mvyB;
				RecY = RecYB;
				ref_frame = ref_frameB;
			}

			if(sub_pos != -1 && ref_frame != -1 ) // if no full-pel motion
			{
				if((sub_pos+1)%4 == 0 || (sub_pos+1)/4 == 0)	    // if line or column, then only 1D-Filter,
					number_equation = FILTER_SIZE;					// hor or vert. resp.
				else
					number_equation = SQR(FILTER_SIZE);

				for(yi = 0; yi < 4; yi++)    //y
				{
					for(xj = 0; xj < 4; xj++)  //x
					{
						for(equation = 0; equation < number_equation; equation++)
						{
							if(number_equation == SQR(FILTER_SIZE))
							{
								filter_pos_x1 = FindPosition(img->width, x+xj,equation%FILTER_SIZE,mvx);					//current derivation (h_ij)
								filter_pos_y1 = FindPosition(img->height,y+yi,equation/FILTER_SIZE,mvy);					//current derivation (h_ij)
								for(i = 0; i < FILTER_SIZE; i++)
								{
									for(j = 0; j < FILTER_SIZE; j++)
									{
										filter_pos_x2 = FindPosition(img->width, x+xj,j,mvx);
										filter_pos_y2 = FindPosition(img->height,y+yi,i,mvy);
										ReorderPosition(new_filter_pos,  new_eq,   new_sub_pos,
																		FILTER_SIZE*i+j, equation, sub_pos);
										NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos] += b_scaling*
																		RecY[filter_pos_y1][filter_pos_x1]*
																		RecY[filter_pos_y2][filter_pos_x2];
									}
								}
							}
							else   // 1D-Filter case
							{
								if((sub_pos+1) / 4 == 0)	// horizontal
								{
									filter_pos_x1 = FindPosition(img->width, x+xj,equation%FILTER_SIZE,mvx); 																							//current derivation (h_ij)
									filter_pos_y1 = FindPosition(img->height,y+yi,FILTER_OFFSET,mvy); 																							//current derivation (h_ij)
									for(i = 0; i < FILTER_SIZE; i++)
									{
										filter_pos_x2 = FindPosition(img->width, x+xj, i,mvx);
										filter_pos_y2 = FindPosition(img->height,y+yi, FILTER_OFFSET,mvy);
										ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
																		i, equation,   sub_pos);

										NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos] += b_scaling*
															RecY[filter_pos_y1][filter_pos_x1]*
															RecY[filter_pos_y2][filter_pos_x2];
									}
								}
								else // vertical
								{
									filter_pos_x1 = FindPosition(img->width, x+xj,FILTER_OFFSET,mvx);																									//current derivation (h_ij)
									filter_pos_y1 = FindPosition(img->height,y+yi,equation%FILTER_SIZE,mvy);																							//current derivation (h_ij)
									for(i = 0; i < FILTER_SIZE; i++)
									{
										filter_pos_x2 = FindPosition(img->width, x+xj, FILTER_OFFSET,mvx);
										filter_pos_y2 = FindPosition(img->height,y+yi, i,mvy);
										ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
																		i, equation,   sub_pos);
										NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos] += b_scaling*
																		RecY[filter_pos_y1][filter_pos_x1]*
																		RecY[filter_pos_y2][filter_pos_x2];
									}
								}
							}
							ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
															1, equation,   sub_pos);   		// 1 is not important
							N_Vector[*new_sub_pos][*new_eq] += b_scaling*imgY_org[y+yi][x+xj]*
														RecY[filter_pos_y1][filter_pos_x1];
						}
					}
				}
			}
		}
	}
}


/*!
 ************************************************************************
 * \brief
 *    sets necessary filter data for current macroblock in 1/8-pel case
 ************************************************************************
 */

void SetAdaptiveFilterEighthpel(void)
{
	Macroblock *currMB = &img->mb_data[img->current_mb_nr];
	int i, j, x, y, xj, yi;
	float b_scaling=1.0;
	int run=0, rerun=0;
	int run8=0, rerun8 = 1;
	int equation=0;
	int mvx_subF=0, mvy_subF=0;
	int mvx_subB=0, mvy_subB=0;
	int mvx=0,  mvy=0;
	int mvxF=0, mvyF=0;
	int mvxB=0, mvyB=0;
	int mvxF8, mvyF8, mvxB8, mvyB8;
	int tmp_x_f, tmp_y_f, tmp_x_b, tmp_y_b;
	int sub_pos=0;
	int sub_posF=0;
	int sub_posB=0;
	int ref_frame=0;
	int ref_frameF=-1;
	int ref_frameB=-1;
	int subblock=0;
	int filter_pos_x1, filter_pos_y1, filter_pos_x2, filter_pos_y2;
	int x_orig, y_orig;       // absolute distances from current macroblock
	int h1, h2, h3;
	int *new_filter_pos, *new_eq, *new_sub_pos;
	int number_equation;
	imgpel **RecY=NULL;
	imgpel **RecYF=NULL;
	imgpel **RecYB=NULL;
	new_filter_pos = &h1;
	new_eq = &h2;
	new_sub_pos = &h3;

	if(IS_INTRA(currMB)) //intra macroblocks are not used for calculation of the filter coeffs.
		return;

	x_orig = MB_BLOCK_SIZE*(img->block_x/4);
	y_orig = MB_BLOCK_SIZE*(img->block_y/4);
	for(run8=0;run8<rerun8;run8++)
	{
		for(subblock = 0; subblock < 16; subblock++)
		{
			x = x_orig+4*(subblock%4);
			y = y_orig+4*(subblock/4);
			mvxF8 = enc_picture->mv[LIST_0][y/4][x/4][0];
			mvyF8 = enc_picture->mv[LIST_0][y/4][x/4][1];
			if(mvxF8>=0)
				tmp_x_f = mvxF8%8;
			else
				tmp_x_f = 8-(abs(mvxF8)%8);
			if(mvyF8>=0)
				tmp_y_f = mvyF8%8;
			else
				tmp_y_f = 8-(abs(mvyF8)%8);
			
			if(run8 == 0)
			{
				if(	abs(mvxF8)%2 == 0  && abs(mvyF8)%2 == 0 )		// quarter-pel
				{
					mvxF = mvxF8/2;
					mvyF = mvyF8/2;
				}
				else if(abs(mvxF8)%2 == 0)		// quarter-pel in x
				{
					mvxF = mvxF8/2;
					mvyF = (mvyF8-1)/2;
				}
				else if(abs(mvyF8)%2 == 0)		// quarter-pel in y
				{
					mvxF = (mvxF8-1)/2;
					mvyF = mvyF8/2;
				}
				else		// upper right - down left
								// or upper left - down right interpolation
				{
					if((tmp_x_f+tmp_y_f)%4 == 0)		// upper right-down left
					{
						mvxF = (mvxF8+1)/2;
						mvyF = (mvyF8-1)/2;
					}
					else		// upper left - down right
					{
						mvxF = (mvxF8-1)/2;
						mvyF = (mvyF8-1)/2;
					}
				}
			}
			else // if(run8 == 0)
			{
				if(	abs(mvxF8)%2 == 0  && abs(mvyF8)%2 == 0 )		// quarter-pel
				{
					mvxF = mvxF8/2;
					mvyF = mvyF8/2;
				}
				else if(abs(mvxF8)%2 == 0)		// quarter-pel in x
				{
					mvxF = mvxF8/2;
					mvyF = (mvyF8+1)/2;
				}
				else if(abs(mvyF8)%2 == 0)		// quarter-pel in y
				{
					mvxF = (mvxF8+1)/2;
					mvyF = mvyF8/2;
				}
				else		// upper right - down left
								// or upper left - down right interpolation
				{
 					if((tmp_x_f+tmp_y_f)%4 == 0)				// upper right - down left interpolation
					{
						mvxF = (mvxF8-1)/2;
						mvyF = (mvyF8+1)/2;
					}
					else						// upper left - down right interpolation
					{
						mvxF = (mvxF8+1)/2;
						mvyF = (mvyF8+1)/2;
					}
				}
			}
			
			if(mvxF >= 0)
				mvx_subF = mvxF%4;      // x-sub-coordinate in a 4x4block
			else
				mvx_subF = (4-abs(mvxF)%4)%4;
			if(mvyF >= 0)
				mvy_subF = mvyF%4;      // y-sub-coordinate in a 4x4block
			else
				mvy_subF = (4-abs(mvyF)%4)%4;
	
			sub_posF = mvx_subF + 4*mvy_subF-1;    // pos 0..14 in a 4x4 block
			ref_frameF = enc_picture->ref_idx[LIST_0][y/4][x/4];
			if(ref_frameF != -1)
				RecYF  = listX[LIST_0][ref_frameF]->imgY;
			if(img->type == B_SLICE)
			{
				mvxB8= enc_picture->mv[LIST_1][y/4][x/4][0];
				mvyB8= enc_picture->mv[LIST_1][y/4][x/4][1];
				if(mvxB8>=0)
					tmp_x_b = mvxB8%8;
				else
					tmp_x_b = 8-(abs(mvxB8)%8);
				if(mvyB8>=0)
					tmp_y_b = mvyB8%8;
				else
					tmp_y_b = 8-(abs(mvyB8)%8);
				if(run8 == 0)
				{
					if(	abs(mvxF8)%2 == 0  && abs(mvyF8)%2 == 0 )		// quarter-pel
					{
						mvxB = mvxB8/2;
						mvyB = mvyB8/2;
					}
					else if(abs(mvxB8)%2 == 0)		// quarter-pel in x
					{
						mvxB = mvxB8/2;
						mvyB = (mvyB8-1)/2;
					}
					else if(abs(mvyB8)%2 == 0)		// quarter-pel in y
					{
						mvxB = (mvxB8-1)/2;
						mvyB = mvyB8/2;
					}
					else		// upper right - down left
									// or upper left - down right interpolation
					{
						if((tmp_x_b+tmp_y_b)%4 == 0)		// upper right-down left
						{
							mvxB = (mvxB8+1)/2;
							mvyB = (mvyB8-1)/2;
						}
						else		// upper left - down right
						{
							mvxB = (mvxB8-1)/2;
							mvyB = (mvyB8-1)/2;
						}
					}
				}
				else
				{
					if(	abs(mvxB8)%2 == 0  && abs(mvyB8)%2 == 0 )		// quarter-pel
					{
						mvxB = mvxB8/2;
						mvyB = mvyB8/2;
					}
					else if(abs(mvxB8)%2 == 0)		// quarter-pel in x
					{
						mvxB = mvxB8/2;
						mvyB = (mvyB8+1)/2;
					}
					else if(abs(mvyB8)%2 == 0)		// quarter-pel in y
					{
						mvxB = (mvxB8+1)/2;
						mvyB = mvyB8/2;
					}
					else		// upper right - down left
									// or upper left - down right interpolation
					{
						if((tmp_x_b+tmp_y_b)%4 == 0)				// upper right - down left interpolation
						{
							mvxB = (mvxB8-1)/2;
							mvyB = (mvyB8+1)/2;
						}
						else						// upper left - down right interpolation
						{
							mvxB = (mvxB8+1)/2;
							mvyB = (mvyB8+1)/2;
						}
					}
				}

				if(mvxB >= 0)
					mvx_subB = mvxB%4;      // x-sub-coordinate in a 4x4block
				else
					mvx_subB = (4-abs(mvxB)%4)%4;
				if(mvyB >= 0)
					mvy_subB = mvyB%4;      // y-sub-coordinate in a 4x4block
				else
					mvy_subB = (4-abs(mvyB)%4)%4;
		
				sub_posB = mvx_subB + 4*mvy_subB-1;    // pos 0..14 in a 4x4 block
				ref_frameB = enc_picture->ref_idx[LIST_1][y/4][x/4];
				if(ref_frameB != -1)
					RecYB  = listX[LIST_1][ref_frameB]->imgY;
				rerun = 1;
			}
			else
				rerun = 0;
			if(ref_frameB != -1 && ref_frameF != -1)
				b_scaling = 0.5;
			else
				b_scaling = 1.0;
			for(run = 0; run <= rerun; run++)
			{
				if(run == 0)
				{
					sub_pos = sub_posF;
					mvx = mvxF;
					mvy = mvyF;
					RecY = RecYF;
					ref_frame = ref_frameF;
				}
				else
				{
					sub_pos = sub_posB;
					mvx = mvxB;
					mvy = mvyB;
					RecY = RecYB;
					ref_frame = ref_frameB;
				}
	
				if(sub_pos != -1 && ref_frame != -1 ) // if no full-pel motion
				{
					if((sub_pos+1)%4 == 0 || (sub_pos+1)/4 == 0)	    // if line or column, then only 1D-Filter,
						number_equation = FILTER_SIZE;					// hor or vert. resp.
					else
						number_equation = SQR(FILTER_SIZE);
	
					for(yi = 0; yi < 4; yi++)    //y
					{
						for(xj = 0; xj < 4; xj++)  //x
						{
							for(equation = 0; equation < number_equation; equation++)
							{
								if(number_equation == SQR(FILTER_SIZE))
								{
									filter_pos_x1 = FindPosition(img->width, x+xj,equation%FILTER_SIZE,mvx);					//current derivation (h_ij)
									filter_pos_y1 = FindPosition(img->height,y+yi,equation/FILTER_SIZE,mvy);					//current derivation (h_ij)
									for(i = 0; i < FILTER_SIZE; i++)
									{
										for(j = 0; j < FILTER_SIZE; j++)
										{
											filter_pos_x2 = FindPosition(img->width, x+xj,j,mvx);
											filter_pos_y2 = FindPosition(img->height,y+yi,i,mvy);
											ReorderPosition(new_filter_pos,  new_eq,   new_sub_pos,
																			FILTER_SIZE*i+j, equation, sub_pos);
											NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos] += 0.5*b_scaling*
																			RecY[filter_pos_y1][filter_pos_x1]*
																			RecY[filter_pos_y2][filter_pos_x2];
										}
									}
								}
								else   // 1D-Filter case
								{
									if((sub_pos+1) / 4 == 0)	// horizontal
									{
										filter_pos_x1 = FindPosition(img->width, x+xj,equation%FILTER_SIZE,mvx); 																							//current derivation (h_ij)
										filter_pos_y1 = FindPosition(img->height,y+yi,FILTER_OFFSET,mvy); 																							//current derivation (h_ij)
										for(i = 0; i < FILTER_SIZE; i++)
										{
											filter_pos_x2 = FindPosition(img->width, x+xj, i,mvx);
											filter_pos_y2 = FindPosition(img->height,y+yi, FILTER_OFFSET,mvy);
											ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
																			i, equation,   sub_pos);
	
											NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos]+= 0.5*b_scaling*RecY[filter_pos_y1][filter_pos_x1]*RecY[filter_pos_y2][filter_pos_x2];
										}
									}
									else // vertical
									{
										filter_pos_x1 = FindPosition(img->width, x+xj,FILTER_OFFSET,mvx);																									//current derivation (h_ij)
										filter_pos_y1 = FindPosition(img->height,y+yi,equation%FILTER_SIZE,mvy);																							//current derivation (h_ij)
										for(i = 0; i < FILTER_SIZE; i++)
										{
											filter_pos_x2 = FindPosition(img->width, x+xj, FILTER_OFFSET,mvx);
											filter_pos_y2 = FindPosition(img->height,y+yi, i,mvy);
											ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
																			i, equation,   sub_pos);
											NxN_Matrix[*new_sub_pos][*new_eq][*new_filter_pos] += 0.5*b_scaling*RecY[filter_pos_y1][filter_pos_x1]*RecY[filter_pos_y2][filter_pos_x2];
										}
									}
								}
								ReorderPosition(new_filter_pos,   new_eq, new_sub_pos,
																1, equation,   sub_pos);   		// 1 is not important
								N_Vector[*new_sub_pos][*new_eq] += 0.5*b_scaling*imgY_org[y+yi][x+xj]*
															RecY[filter_pos_y1][filter_pos_x1];
							}
						}
					}
				}
			}
		}
	}
}

/*!
 ************************************************************************
 * \brief
 *    finds current position in X-direction
 *    input:
 *      DimX   - X-size of image,
 *      curX   - current X-coordinate,
 *      Offset - offset in X-direction from left filter position (2D-filter)
 *      mvx    - x-motion vector
 ************************************************************************
 */
int FindPosition(int Dim, int cur_pos, int Offset, int mv)
{
	int position;
	if(mv < 0 && abs(mv)%4!=0)
		mv = mv-4;
	mv =mv/4;
	position = cur_pos-FILTER_OFFSET+Offset+mv;
	return max(0, min(Dim-1, position ) );
}


/*!
 ************************************************************************
 * \brief
 *    reorder position, depending on current sub-position
 *    in order to guarantee symmetric filter
 *    input:
 *      cur_xy   - current unsymmetric filter position
 *      equation - current filter position
 *      sub_pos  - sub-position, depending on mvx and mvy
 *    output:
 *      *new_filter_pos   - new symmetric filter position
 *      *new_eq   - new equation
 *      *new_sub_pos - new sub_position in a symmetric filter
 ************************************************************************
 */

void ReorderPosition(int *new_filter_pos, int *new_eq,  int *new_sub_pos,
										 int cur_xy,          int equation, int sub_pos)
{
	*new_sub_pos    = FILTER_NEW_SUB_POS[sub_pos];
	*new_eq         = TwoDEquationPattern[sub_pos][equation];
	*new_filter_pos = TwoDEquationPattern[sub_pos][cur_xy];
}

/*!
 ************************************************************************
 * \brief
 *    extend the filter to 2D (FILTER_SIZE X FILTER_SIZE)
 *
 *    input:
 *      sub_pos - sub-position
 *		 input output:
 *			 FCoef[][] - Filter to be extended at symmetric positions,
 *			 							using FILTER_NEW_SUB_POS
 ************************************************************************
 */
void ExtendFilterCoefficientsDouble(int sub_pos, double FCoef[15][SQR_FILTER])
{
	int i;
	if(SymmetryPosition[sub_pos] == 0)		//check if it is allowed to change original filter coefficients
	{
		for(i = 0; i < SQR_FILTER; i++)
			FCoef[sub_pos][i] = FCoef[FILTER_NEW_SUB_POS[sub_pos]][TwoDSymmetricPattern[sub_pos][i]];
	}
}


/*!
 ************************************************************************
 * \brief
 *    extend the filter to 2D (FILTER_SIZE X FILTER_SIZE)
 *
 *    input:
 *      sub_pos - sub-position
 *		 input output:
 *			 FCoef[][] - Filter to be extended at symmetric positions,
 *			 							using FILTER_NEW_SUB_POS
 ************************************************************************
 */
void ExtendFilterCoefficientsInt(int sub_pos, int ICoef[15][SQR_FILTER])
{
	int i;
	if(SymmetryPosition[sub_pos] == 0)		//check if it is allowed to change original filter coefficients
	{
		for(i = 0; i < SQR_FILTER; i++)
			ICoef[sub_pos][i] = ICoef[FILTER_NEW_SUB_POS[sub_pos]][TwoDSymmetricPattern[sub_pos][i]];
	}
}


/*!
 ************************************************************************
 * \brief

 *			adapt 1D filter to 2D filter by inserting zeros in 1st,2nd,4th,5th and 6th raw or column
 *			and copying the 3d one
 *
 *    input:
 *      sub_pos - sub-position
 *		 input output:
 *			 FCoef[][] - Filter to be extended at symmetric positions,
 *			 							using FILTER_NEW_SUB_POS
 ************************************************************************
 */
void Position1DFilter(int sub_pos, double FCoef[SQR_FILTER])
{
	int i;
	if((sub_pos+1) % 4 == 0)     // vertical
	{
		for(i = SQR(FILTER_SIZE)-1; i>=0; i--)
			if(i%FILTER_SIZE == FILTER_OFFSET)
				FCoef[i]=FCoef[i/FILTER_SIZE];
		else
			FCoef[i]=0.0;
	}
	else if((sub_pos+1) / 4 == 0)     // horizontal
	{
		for(i = SQR(FILTER_SIZE)-1; i>=0; i--)
			if(i/FILTER_SIZE == FILTER_OFFSET)
				FCoef[i]=FCoef[i%FILTER_SIZE];
		else
			FCoef[i]=0.0;
	}
	else
		return;
}


/*!
 ************************************************************************
 * \brief
 *    finds filter coefficients, solving the equation
 *                      NxN_Matrix*FilterCoef=N_Vector
 ************************************************************************
 */
void FindFilterCoef(void)
{
	int i, j, k;
	double **alpha, **beta;
	double *y;
	int N = SQR(FILTER_SIZE);
	int sub_pos;
	if(!UseAdaptiveFilterForCurrentFrame())
		return;
	get_mem1Ddouble(&y,N);
	get_mem2Ddouble(&alpha,N,N);
	get_mem2Ddouble(&beta,N,N);
	for(sub_pos = a_pos; sub_pos <= o_pos; sub_pos++)
	{
		if(!SymmetryPosition[sub_pos])
		{
			if((FilterFlag[sub_pos] = FilterFlag[FILTER_NEW_SUB_POS[sub_pos]]))
      {
			 if(Is1DPosition[sub_pos])
         for(i = 0; i < FILTER_SIZE; i++)
          {
            FilterCoef[sub_pos][i]=
              CalculatedFilter[FILTER_NEW_SUB_POS[sub_pos]][TwoDEquationPattern[sub_pos][i]];
          }
        else
          for(i = 0; i < SQR_FILTER; i++)
          {
            FilterCoef[sub_pos][i]=
						  CalculatedFilter[FILTER_NEW_SUB_POS[sub_pos]][TwoDEquationPattern[sub_pos][i]];
          }
      }
      else
        for(i = 0; i < SQR_FILTER; i++)
          FilterCoef[sub_pos][i]= STANDARD_2D_FILTER[sub_pos][i];

			continue;
		}

		N = POS_EQUATION_NUMBER[sub_pos];
		FilterFlag[sub_pos]=1;
//1.Step : LU Decomposition

//initialize the ii-components
		for(j = 0; j < N; j++)
		{
			for(i = 0; i < N; i++)
			{
				beta[i][j] = 0.0;
				alpha[i][j] = i==j ? 1.0:0.0;
			}
			y[j] = 0.0;
			CalculatedFilter[sub_pos][j] = 0.0;
		}
		for(j = 0; j < N; j++)
		{
			for(i = 0; i <=j; i++)
			{
				beta[i][j] = NxN_Matrix[sub_pos][i][j];
				for(k = 0; k < i; k++)
				{
					beta[i][j] -= alpha[i][k]*beta[k][j];
				}
			}
			if(fabs(beta[j][j])<0.000001)
				FilterFlag[sub_pos]=0;
			else
			{
				for(i = j+1; i < N; i++)
				{
					alpha[i][j] = NxN_Matrix[sub_pos][i][j];
					for(k = 0; k < j; k++)
					{
						alpha[i][j] -= alpha[i][k]*beta[k][j];
					}
					alpha[i][j] = alpha[i][j]/beta[j][j];
				}
			}
		}
/*
		for(i = 0; i < N; i++)
		for(j = 0; j < N; j++)
		A_decompose[i][j] = (i<=j ? beta[i][j]:alpha[i][j]);
*/
		if(FilterFlag[sub_pos]==0)
		{
      for(i = 0; i < SQR_FILTER; i++)
        FilterCoef[sub_pos][i]= STANDARD_2D_FILTER[sub_pos][i];
			continue;
		}
  //2. Step: Solving L*y = b, forward substitution
		y[0]=N_Vector[sub_pos][0];
		for(i = 1; i < N; i++)
		{
			y[i] = N_Vector[sub_pos][i];
			for(j = 0; j < i; j++)
			{
				y[i] -= alpha[i][j]*y[j];
			}
		}
  //3. Step: Solving U*x = y, backsubstitution
		CalculatedFilter[sub_pos][N-1] = y[N-1]/beta[N-1][N-1];
		for(i = N-2; i >= 0; i--)
		{
			CalculatedFilter[sub_pos][i] = y[i];
			for(j = i+1; j < N; j++)
			{
				CalculatedFilter[sub_pos][i] -= beta[i][j]*CalculatedFilter[sub_pos][j];
			}
			CalculatedFilter[sub_pos][i] = CalculatedFilter[sub_pos][i]/beta[i][i];
		}
		for(i = 0; i < SQR_FILTER; i++)
			FilterCoef[sub_pos][i]=
					CalculatedFilter[FILTER_NEW_SUB_POS[sub_pos]][TwoDEquationPattern[sub_pos][i]];
	}
  // positioning 1D filter
	for(sub_pos = a_pos; sub_pos <= o_pos; sub_pos++)
	{
    if(FilterFlag[sub_pos])
		  Position1DFilter(sub_pos, FilterCoef[sub_pos]);
	}
	// calculate UseAllPositions
	SubpelPositionsPattern = 	(FilterFlag[j_pos]<<4) + (FilterFlag[f_pos]<<3) + (FilterFlag[e_pos]<<2) + 
			(FilterFlag[b_pos]<<1) +  FilterFlag[a_pos];
	if( SubpelPositionsPattern == 31)
		UseAllSubpelPositions = 1;					// for indicating that every sub-pel position is used
	else
		UseAllSubpelPositions	= 0;


	PredictFilterCoefficients();

//	PrintFilterCoefInt(DiffQFilterCoef);
//	PrintFilterCoefDouble(PredFilterCoef);
//	PrintFilterCoefDouble(FilterCoef);

	free_mem1Ddouble(y);
	free_mem2Ddouble(alpha);
	free_mem2Ddouble(beta);
}

/*!
 ************************************************************************
 * \brief
 *		quantize filter coefficients
 *		input:
 *				double FCoef[][] - filter coefficients to be quantized
 *		output:
 *				int QFCoef[][]   - quantized filter coefficients, 
 *														 represented in step numbers
 ************************************************************************
 */
void QuantizeFilterCoefficients(double FCoef[SQR_FILTER], int QFCoef[SQR_FILTER])
{
	double number_of_steps = pow(2,NumberOfQBits-1);  // 1 bit for sign, NumberOfQBits-1 for amplitude
	int sign;
	int j;
	for(j=0; j < SQR_FILTER; j++)
	{
		sign = FCoef[j] >= 0.0 ? 1:-1;
		QFCoef[j] = sign*(int)(fabs(FCoef[j])*number_of_steps + 0.5);		// step number
		FCoef[j] = (double)(QFCoef[j])/number_of_steps; // reconstructed (quantized) filter coefficients
	}
}

/*!
 ************************************************************************
 * \brief
 *    predict filter coefficients
 *
 ************************************************************************
 */
void PredictFilterCoefficients(void)
{
	int i,j;
// at first predict b and h position using inter prediction, b and h predicted values are already quantized in last frame
	for(i=0; i < SQR_FILTER; i++)
		PredFilterCoef[b_pos][i] = STANDARD_2D_FILTER[b_pos][i];

	for(i = 0; i < SQR_FILTER; i++)
		DiffFilterCoef[b_pos][i] = FilterCoef[b_pos][i] - PredFilterCoef[b_pos][i];
	QuantizeFilterCoefficients(DiffFilterCoef[b_pos], DiffQFilterCoef[b_pos]);

	for(i = 0; i < SQR_FILTER; i++)
		FilterCoef[b_pos][i] = DiffFilterCoef[b_pos][i] + PredFilterCoef[b_pos][i];
	
	ExtendFilterCoefficientsDouble(h_pos, FilterCoef);
	
	// predict a using inter-prediction

	for(i=0; i < SQR_FILTER; i++)
		PredFilterCoef[a_pos][i] = STANDARD_2D_FILTER[a_pos][i];

	for(i = 0; i < SQR_FILTER; i++)
		DiffFilterCoef[a_pos][i] = FilterCoef[a_pos][i] - PredFilterCoef[a_pos][i];
	QuantizeFilterCoefficients(DiffFilterCoef[a_pos], DiffQFilterCoef[a_pos]);
	for(i = 0; i < SQR_FILTER; i++)
		FilterCoef[a_pos][i] = DiffFilterCoef[a_pos][i] + PredFilterCoef[a_pos][i];

	ExtendFilterCoefficientsDouble(c_pos, FilterCoef);
	ExtendFilterCoefficientsDouble(d_pos, FilterCoef);
	ExtendFilterCoefficientsDouble(l_pos, FilterCoef);
	
	// predict e_pos from a_pos and d_pos by multiplication
	// predict f_pos from b_pos and d_pos by multiplication
	// predict j_pos from b_pos and h_pos by multiplication
  for(i = 0; i < FILTER_SIZE; i++)
  {
		for(j = 0; j < FILTER_SIZE; j++)
		{
      PredFilterCoef[e_pos][FILTER_SIZE*i+j] =
				FilterCoef[d_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[a_pos][FILTER_SIZE*FILTER_OFFSET+j];
      PredFilterCoef[f_pos][FILTER_SIZE*i+j] =
				  FilterCoef[d_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[b_pos][FILTER_SIZE*FILTER_OFFSET+j];
      PredFilterCoef[j_pos][FILTER_SIZE*i+j] =
        FilterCoef[h_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[b_pos][FILTER_SIZE*FILTER_OFFSET+j];
    }
  }
  for(i = 0; i < SQR_FILTER; i++)
  {
    if(FilterFlag[e_pos])
      DiffFilterCoef[e_pos][i] = FilterCoef[e_pos][i] - PredFilterCoef[e_pos][i];
    if(FilterFlag[f_pos])
      DiffFilterCoef[f_pos][i] = FilterCoef[f_pos][i] - PredFilterCoef[f_pos][i];
    if(FilterFlag[j_pos])
      DiffFilterCoef[j_pos][i] = FilterCoef[j_pos][i] - PredFilterCoef[j_pos][i];
  }
  if(FilterFlag[e_pos])
    QuantizeFilterCoefficients(DiffFilterCoef[e_pos], DiffQFilterCoef[e_pos]);
  if(FilterFlag[f_pos])
    QuantizeFilterCoefficients(DiffFilterCoef[f_pos], DiffQFilterCoef[f_pos]);
  if(FilterFlag[j_pos])
    QuantizeFilterCoefficients(DiffFilterCoef[j_pos], DiffQFilterCoef[j_pos]);
  for(i = 0; i < SQR_FILTER; i++)
  {
    if(FilterFlag[e_pos])
      FilterCoef[e_pos][i] = DiffFilterCoef[e_pos][i] + PredFilterCoef[e_pos][i];
    if(FilterFlag[f_pos])
      FilterCoef[f_pos][i] = DiffFilterCoef[f_pos][i] + PredFilterCoef[f_pos][i];
    if(FilterFlag[j_pos])
      FilterCoef[j_pos][i] = DiffFilterCoef[j_pos][i] + PredFilterCoef[j_pos][i];
  }
  if(FilterFlag[e_pos])
  {
    ExtendFilterCoefficientsDouble(g_pos, FilterCoef);
    ExtendFilterCoefficientsDouble(m_pos, FilterCoef);
    ExtendFilterCoefficientsDouble(o_pos, FilterCoef);
  }
  if(FilterFlag[f_pos])
  {
    ExtendFilterCoefficientsDouble(i_pos, FilterCoef);
    ExtendFilterCoefficientsDouble(k_pos, FilterCoef);
    ExtendFilterCoefficientsDouble(n_pos, FilterCoef);
  }
}


/*!
 ************************************************************************
 * \brief
 *    Upsample 4 times with new calculated filter coefficients.
 *    Store the old upsampled frames.
 ************************************************************************
 */
void UnifiedOneForthPixWithNewFilter(void)
{

	int i, j,ii, jj;
	unsigned frame;
	imgpel **out4Y_aif, **out4Y;
	imgpel **refY;
	int img_width, img_height;
	int sub_pos, pos_x, pos_y, x_sub, y_sub;
	double is;
	if(current_poc == img->ThisPOC)
		return;			// avoids multiple interolating of the same frames
	else
		current_poc = img->ThisPOC;
	for(frame = 0; frame < dpb.ref_frames_in_buffer; frame++)
	{
		img_width  = dpb.fs_ref[frame]->frame->size_x;
		img_height = dpb.fs_ref[frame]->frame->size_y;
		out4Y_aif  = dpb.fs_ref[frame]->frame->imgY_ups_aif;
		out4Y      = dpb.fs_ref[frame]->frame->imgY_ups;
		refY       = dpb.fs_ref[frame]->frame->imgY;

		for(i=-4*IMG_PAD_SIZE; i < (IMG_PAD_SIZE+img_height)*4; i++)
			for(j=-4*IMG_PAD_SIZE; j < (IMG_PAD_SIZE+img_width)*4; j++)
		{
			x_sub = (j+4*IMG_PAD_SIZE)%4;       // x-sub-coordinate in a 4x4block
			y_sub = (i+4*IMG_PAD_SIZE)%4;       // y-sub-coordinate in a 4x4block
			sub_pos = x_sub + 4*y_sub;          // pos 1..15 in a 4x4 block
			is = 0.0;
			if(sub_pos)
			{
				for(ii = 0; ii < FILTER_SIZE; ii++)
					for(jj = 0; jj < FILTER_SIZE; jj++)
				{
					pos_y = max(0, min(img_height-1, i/4-FILTER_OFFSET+ii ) );
					if(i<0 && pos_y > 0)
						pos_y -=1;
					pos_x = max(0, min(img_width -1, j/4-FILTER_OFFSET+jj ) );
					if(j<0 && pos_x > 0)
						pos_x -=1;
					is += (FilterCoef[sub_pos-1][FILTER_SIZE*ii+jj]*refY[pos_y][pos_x]);
//						is += (STANDARD_2D_FILTER[sub_pos-1][FILTER_SIZE*ii+jj]*refY[pos_y][pos_x]);
				}
				out4Y_aif[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] = max(0,min(255,(int)(is+0.5)));
			}
			else
				out4Y_aif[i+4*IMG_PAD_SIZE][j+4*IMG_PAD_SIZE] =
						refY[max(0,min(img_height-1,i/4))][max(0,min(img_width-1,j/4))];
		}
	}
//	FindSAD();
}

/*!
 ************************************************************************
 * \brief
 *    swap standard and new calculated
 *    interpolated upsamled frames
 ************************************************************************
 */
void SwapUpsampledFrames(void)
{
	unsigned frame;
	
	for(frame = 0; frame < dpb.ref_frames_in_buffer; frame++)
	{
		dpb.fs_ref[frame]->frame->p_imgY_ups_aif = dpb.fs_ref[frame]->frame->imgY_ups_aif;
		dpb.fs_ref[frame]->frame->imgY_ups_aif = dpb.fs_ref[frame]->frame->imgY_ups;
		dpb.fs_ref[frame]->frame->imgY_ups = dpb.fs_ref[frame]->frame->p_imgY_ups_aif;
	}
}

/*!
 ************************************************************************
 * \brief
 *    store filter coefficients for current reference frame
 ************************************************************************
 */
void StoreFilterCoefficients(StorablePicture *p)
{
	int sub_pel, i;
	if(UseAdaptiveFilterForCurrentFrame() && img->AdaptiveFilterFlag == 1)
		for(sub_pel = 0; sub_pel < 15; sub_pel++)
			for(i = 0; i < SQR_FILTER; i++)
				p->FilterCoefficients[sub_pel][i] = FilterCoef[sub_pel][i];
	else
		for(sub_pel = 0; sub_pel < 15; sub_pel++)
			for(i = 0; i < SQR_FILTER; i++)
				p->FilterCoefficients[sub_pel][i] = STANDARD_2D_FILTER[sub_pel][i];
}



/*!
 ************************************************************************
 * \brief
 *    prints filter coefficient
 ************************************************************************
 */
void PrintFilterCoefDouble(double FCoef[15][SQR_FILTER])
{
	int i,j;
	for(i = a_pos; i <= o_pos; i++)
	{
//		printf("Filter_%c=[\n",97+i);
    printf("FilterFlag[%c]=%d\n",97+i,FilterFlag[i]);
		printf("Filter(%d,1:6,1:6)=[\n",i+1);
		for(j = 0; j < SQR(FILTER_SIZE); j++)
		{
			if(!(j%FILTER_SIZE))
				printf("[");
			printf("%12.8f   ",FCoef[i][j]);
			if(j%FILTER_SIZE == FILTER_SIZE-1)
				printf("];\n");
			else
				printf(", ");
		}
		printf("];\n");
	}
}

/*!
 ************************************************************************
 * \brief
 *    prints filter coefficients
 ************************************************************************
 */
void PrintFilterCoefInt(int FCoef[15][SQR_FILTER])
{
	int i,j;
	for(i = a_pos; i <= o_pos; i++)
	{
		printf("%c_pos:\n",97+i);
		for(j = 0; j < SQR(FILTER_SIZE); j++)
		{
			if(FCoef[i][j] >= 0)
				printf(" ");
			printf("%d   \t",FCoef[i][j]);
			if(j%FILTER_SIZE == 5)
				printf("\n");
		}
		printf("========================================================\n");
	}
}

#endif
